/*
	CVSNT Generic API
    Copyright (C) 2004 Tony Hoyle and March-Hare Software Ltd

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
/* Unix specific */
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif
#include <system.h>

#include <sys/types.h>
#include <sys/wait.h>

#include <signal.h>

#include "../RunFile.h"
#include "../ServerIO.h"

enum enPipes
{
	PIPE_READ = 0,
	PIPE_WRITE = 1 
};

CRunFile::CRunFile()
{
	m_inputFn=NULL;
	m_outputFn=NULL;
	m_errorFn=NULL;
	m_child = 0;

	/* It's impossible to generically get the current executable filename
	   in Unix (see google for much discussion of this), so we set it to 'cvs'
	   and hope it's in the path. */
	program_filename="cvsnt";
}

CRunFile::~CRunFile()
{
}

bool CRunFile::resetArgs()
{
	return m_args.resetArgs();
}

bool CRunFile::setArgs(const char *args)
{
	return m_args.setArgs(args);
}


bool CRunFile::setArgs(int argc, const char *const*argv)
{
	return m_args.setArgs(argc, argv);
}

bool CRunFile::addArg(const char *arg)
{
	return m_args.addArg(arg);
}

bool CRunFile::setInput(int (*inputFn)(char *,size_t))
{
	m_inputFn = inputFn;
	return true;
}

bool CRunFile::setOutput(int (*outputFn)(const char *,size_t))
{
	m_outputFn = outputFn;
	return true;
}

bool CRunFile::setError(int (*errorFn)(const char *,size_t))
{
	m_errorFn = errorFn;
	return true;
}

bool CRunFile::run(const char *path)
{
  	int pid;
  	int fd1[2],fd2[2],fd3[2];

  	if(m_inputFn)
  	{
                pipe(fd1);
                m_inFd=fd1[PIPE_WRITE];
        }
        else
                m_inFd=-1;

        if(m_outputFn)
        {
                pipe(fd2);
                m_outFd=fd2[PIPE_READ];
        }
        else
                m_outFd=-1;

        if(m_errorFn)
        {
                pipe(fd3);
                m_errFd=fd3[PIPE_READ];
        }
        else
        {
                if(m_outFd>=0)
                {
                        m_errFd=dup(m_outFd);
                        fd3[PIPE_WRITE]=-1;
                }
		else
			m_errFd=-1;
	}

	if(path)
	  m_args.insertArg(0,path);

	pid = fork();
	if(pid<0)
		return false;
	
	signal(SIGPIPE,SIG_IGN);
	if(!pid) /* Child */
	{	
		close(fd1[PIPE_WRITE]);
		dup2(fd1[PIPE_READ],0);
		close(fd2[PIPE_READ]);
		dup2(fd2[PIPE_WRITE],1);
		close(fd2[PIPE_READ]);
		dup2(fd2[PIPE_WRITE],2);

		char **myargv = (char**)m_args.toArgv();
		execvp (myargv[0], myargv);
		perror("Exec failed");
		exit(-1);
	}

	close(fd1[PIPE_READ]);
	close(fd2[PIPE_WRITE]);
	close(fd3[PIPE_WRITE]);

	m_child = pid;

	return true;
}

/* Unix gets no I/O until wait() is called - if we had
   reliable threads this wouldn't be necessary... fork()
   isolates the address spaces so is not suitable for this
   task */
bool CRunFile::wait(int& result, int timeout)
{	
	char buf[BUFSIZ],inbuf[BUFSIZ];
	int status,size,wsize,w,l;

	CServerIo::trace(3,"wait() called, m_child=%d",m_child);
	
	if(!m_child)
	  return -1;
		
	if(m_outFd>=0)
		fcntl(m_outFd,F_SETFL, O_NONBLOCK);
	if(m_errFd>=0)
    		fcntl(m_errFd,F_SETFL, O_NONBLOCK);
	if(m_inFd>=0)
    		fcntl(m_errFd,F_SETFL, O_NONBLOCK);

	size = 0;
	if(m_inFd>=0)
	{
		size=l=m_inputFn(inbuf,sizeof(inbuf));
		if(size<=0)
		{
			close(m_inFd);
			m_inFd=-1;
		}
	}
	w = waitpid(m_child,&status,WNOHANG);
	while((timeout==-1 || timeout>0) && ((m_inFd>=0 && size>0) || m_outFd>=0 || m_errFd>=0) && !w)
	{
		if(m_inFd>=0 && size>0)
		{
		  int s = write(m_inFd, inbuf+l-size,size);
		  if(s<0)
		  {
		    close(m_inFd);
		    m_inFd=-1;
		  }
		  if(s)
		  {
		    size-=s;
		    if(!size)
		      size=l=m_inputFn(inbuf,sizeof(inbuf));
		    if(size<=0)
		    {
			close(m_inFd);
			m_inFd=-1;
		    }
		  }
		}
		if(!w)
			w = waitpid(m_child,&status,WNOHANG);
		wsize=0;
      		while(!w && m_errFd>=0 && (wsize=read(m_errFd,buf,sizeof(buf)))>0)
		{
       			(m_errorFn?m_errorFn:m_outputFn)(buf,wsize);
		}
		if(wsize<0 && errno!=EAGAIN)
		{
			close(m_errFd);
			m_errFd=-1;
		}
		if(!w)
			w = waitpid(m_child,&status,WNOHANG);
		wsize=0;
      		while(!w && m_outFd>=0 && (wsize=read(m_outFd,buf,sizeof(buf)))>0)
		{
       			m_outputFn(buf,wsize);
		}
		if(wsize<0 && errno!=EAGAIN)
		{
			close(m_outFd);
			m_outFd=-1;
		}
		if(!w)
		{
		  usleep(100);
		  w = waitpid(m_child,&status,WNOHANG);
		}
		if(timeout!=-1)
		{
			timeout-=100;
			if(timeout==-1) timeout--;
		}
	}
	if(!w && timeout!=-1 && timeout<=0) /* timed out */
	   return false;
	if(m_inFd)
	{
		close(m_inFd);
		m_inFd=-1;
	}

	/* If we've fallen through all of the above with the app still running, wait for it */
	if(!w)
	{
		if(timeout==-1)
			waitpid(m_child,&status,0);
		else
		{
			while(!w && timeout>0)
			{
			   w=waitpid(m_child,&status,WNOHANG);
		 	   usleep(100);
			   timeout-=100;
			}
			if(!w)
			   return false;
		}
	}     	
	else /* App died.. soak up stdio */
	{
		CServerIo::trace(3,"Process finished");
      		while(m_errFd>=0 && (wsize=read(m_errFd,buf,sizeof(buf)))>0)
		{
       			(m_errorFn?m_errorFn:m_outputFn)(buf,wsize);
		}
      		while(m_outFd>=0 && (wsize=read(m_outFd,buf,sizeof(buf)))>0)
		{
       			m_outputFn(buf,wsize);
		}
		if(m_outFd>=0)
		{
			close(m_outFd);	
			m_outFd=-1;
		}
		if(m_errFd>=0)
		{
			close(m_errFd);
			m_errFd=-1;
		}
	}

	m_exitCode=result=WEXITSTATUS(status);
	CServerIo::trace(3,"Exit status is %d",m_exitCode);
	return true;
}
