/*
	CVSNT Generic API
    Copyright (C) 2004 Tony Hoyle and March-Hare Software Ltd

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
/* Win32 specific */
#define WIN32_LEAN_AND_MEAN
#define STRICT
#include <windows.h>

#include <config.h>
#include "../lib/api_system.h"

#include <stdio.h>

#include "../FileAccess.h"

bool CFileAccess::m_bUtf8Mode = false;

CFileAccess::CFileAccess()
{
	m_file = NULL;
}

CFileAccess::~CFileAccess()
{
	close();
}

bool CFileAccess::open(const char *filename, const char *mode)
{
	m_file = _wfopen(Win32Wide(filename),Win32Wide(mode));
	if(!m_file)
		return false;
	return true;
}

bool CFileAccess::open(FILE *file)
{
	if(m_file)
		return false;
	m_file = file;
	return true;
}

bool CFileAccess::close()
{
	if(m_file)
		fclose(m_file);
	m_file = NULL;
	return true;
}

bool CFileAccess::getline(cvs::string& line)
{
	int c;

	if(!m_file)
		return false;

	line.reserve(256);
	line="";

	while((c=fgetc(m_file))!=EOF)
	{
		if(c=='\n')
			break;
		if(c=='\r')
			continue;
		line.append(1,c);
	}
	if(c==EOF && line.empty())
		return false;
	return true;
}

bool CFileAccess::getline(char *line, size_t length)
{
	int c;
	size_t len = length;

	if(!m_file)
		return false;

	while(len && (c=fgetc(m_file))!=EOF)
	{
		if(c=='\n')
			break;
		*(line++)=(char)c;
		--len;
	}
	if(c==EOF && len==length)
		return false;
	return true;
}

bool CFileAccess::putline(const char *line)
{
	if(!m_file)
		return false;

	if(fwrite(line,1,strlen(line),m_file)<strlen(line))
		return false;
	if(fwrite("\n",1,1,m_file)<1)
		return false;
	return true;
}

size_t CFileAccess::read(void *buf, size_t length)
{
	if(!m_file)
		return 0;

	return fread(buf,1,length,m_file);
}

size_t CFileAccess::write(const void *buf, size_t length)
{
	if(!m_file)
		return 0;

	return fwrite(buf,length,1,m_file);
}

loff_t CFileAccess::length()
{
	if(!m_file)
		return 0;

	fpos_t pos,len;

	if(fgetpos(m_file,&pos)<0)
		return 0;
	fseek(m_file,0,SEEK_END);
	if(fgetpos(m_file,&len)<0)
		return 0;
	if(fsetpos(m_file,&pos)<0)
		return 0;
	return (loff_t)len;
}

loff_t CFileAccess::pos()
{
	if(!m_file)
		return 0;

	fpos_t o;
	if(fgetpos(m_file,&o)<0)
		return 0;
	return (loff_t)o;
}

bool CFileAccess::eof()
{
	if(!m_file)
		return false;

	return feof(m_file)?true:false;
}

bool CFileAccess::seek(loff_t pos, SeekEnum whence)
{
	if(!m_file)
		return false;

	fpos_t p;

	switch(whence)
	{
	case seekBegin:
		p=(fpos_t)pos;
		break;
	case seekCurrent:
		if(fgetpos(m_file,&p)<0)
			return false;
		p+=pos;
		break;
	case seekEnd:
		if(pos>0x7FFFFFFF)
		{
			if(fseek(m_file,0x7FFFFFFF,whence)<0)
				return false;
			pos-=0x7FFFFFFFF;
			while(pos>0x7FFFFFFF && !fseek(m_file,-0x7FFFFFFF,SEEK_CUR))
				pos-=0x7FFFFFFFF;
			if(pos>0x7FFFFFFF)
				return false;
			pos=-pos;
		}
		if(fseek(m_file,(long)pos,SEEK_CUR)<0)
			return false;
		return true;
	default:
		return false;
	}
	if(fsetpos(m_file,&p)<0)
		return false;
	return true;
}

cvs::string CFileAccess::tempdir()
{
	cvs::string dir;
	DWORD fa;

	dir.resize(MAX_PATH);
	if(!(GetEnvironmentVariableA("TEMP",(char*)dir.data(),(DWORD)dir.size()) ||
	     GetEnvironmentVariableA("TMP",(char*)dir.data(),(DWORD)dir.size())))
	{
		// No TEMP or TMP, use default <windir>\TEMP
		GetWindowsDirectoryA((char*)dir.data(),(UINT)dir.size());
		strcat((char*)dir.data(),"\\TEMP");
	}
	dir.resize(strlen((char*)dir.data()));
	if((fa=GetFileAttributesA(dir.c_str()))==0xFFFFFFFF || !(fa&FILE_ATTRIBUTE_DIRECTORY))
	{
		// Last resort, can't find a valid temp.... use C:\...
		dir="C:\\";
	}
	return dir;
}

cvs::string CFileAccess::tempfilename(const char *prefix)
{
	cvs::string tempfile;
	tempfile.resize(MAX_PATH);
	GetTempFileNameA(tempdir().c_str(),prefix,0,(char*)tempfile.data());
	tempfile.resize(strlen((char*)tempfile.data()));
	return tempfile;
}

bool CFileAccess::remove(const char *file)
{
	DWORD att = GetFileAttributesA(file);
	if(att==0xFFFFFFFF)
		return false;
	SetFileAttributesA(file,att&~FILE_ATTRIBUTE_READONLY);
	if(!DeleteFileA(file))
		return false;
	return true;
}

void CFileAccess::Win32SetUtf8Mode(bool bUtf8Mode)
{
	m_bUtf8Mode = bUtf8Mode;
}

CFileAccess::Win32Wide::Win32Wide(const char *fn)
{
	if(!fn)
	{
		pbuf=NULL;
		return;
	}
	size_t l = strlen(fn)+1;
	if(l>(sizeof(buf)/sizeof(wchar_t)))
		pbuf=new wchar_t[l];
	else
		pbuf=buf;
	MultiByteToWideChar(m_bUtf8Mode?CP_UTF8:CP_ACP,0,fn,(int)l,pbuf,(int)l*sizeof(wchar_t));
}

CFileAccess::Win32Wide::~Win32Wide()
{
	if(pbuf && pbuf!=buf)
		delete[] pbuf;
}

CFileAccess::Win32Narrow::Win32Narrow(const wchar_t *fn)
{
	if(!fn)
	{
		pbuf=NULL;
		return;
	}
	size_t l = wcslen(fn)+1;
	if((l*3)>sizeof(buf))
		pbuf=new char[l*3];
	else
		pbuf=buf;
	WideCharToMultiByte(m_bUtf8Mode?CP_UTF8:CP_ACP,0,fn,(int)l,pbuf,(int)(l*3),NULL,NULL);
}

CFileAccess::Win32Narrow::~Win32Narrow()
{
	if(pbuf && pbuf!=buf)
		delete[] pbuf;
}

bool CFileAccess::exists(const char *file)
{
	if(GetFileAttributesW(Win32Wide(file))==0xFFFFFFFF)
		return false;
	return true;
}

CFileAccess::TypeEnum CFileAccess::type(const char *file)
{
	DWORD dw = GetFileAttributesW(Win32Wide(file));
	if(dw==0xFFFFFFFF)
		return typeNone;
	if(dw&FILE_ATTRIBUTE_DIRECTORY)
		return typeDirectory;
	return typeFile;
}

bool CFileAccess::absolute(const char *file)
{
	return file[0]=='\\' || file[0]=='/' || file[1]==':';
}

int CFileAccess::uplevel(const char *file)
{
	int level = 0;
	for(const char *p=file; *p; p++)
	{
		size_t l=strcspn(p,"\\/");
		if(l==1 && *p=='.')
			level++; // Compensate for ./
		else if(l==2 && *p=='.' && *(p+1)=='.')
			level+=2; // Compensate for ../
		p+=l;
		if(*p) p++; 
		level--;
	}
	return level;
}

