/* CVS auth library interface

    Copyright (C) 2004 Tony Hoyle and March-Hare Software Ltd

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
#include <config.h>
#include <stdio.h>
#include <ctype.h>
#include <sys/stat.h>
#include <sys/types.h>
#include "../cvsapi/cvsapi.h"

#ifdef USE_SHARED_LIBS
#include <ltdl.h>
#include <glob.h>
#endif

#include <pwd.h>
#include <sys/types.h>
#include <unistd.h>

#ifdef HAVE_SYSLOG_H
#include <syslog.h>
#endif

#include "protocol_interface.h"
#include "library.h"

static const char *_cvs_library_dir_default = CVS_LIBRARY_DIR;
static const char *_cvs_config_dir_default = CVS_CONFIG_DIR;
static const char *cvs_library_dir = _cvs_library_dir_default;
static const char *cvs_config_dir = _cvs_config_dir_default;

static int builtin_yesno_prompt(const char *message, const char *title, int withcancel);
static const char *builtin_get_password(const char *prompt);

static int (*get_cached_password)(const char *key, char *buffer, int buffer_len);
static int (*yesno_prompt)(const char *message, const char *title, int withcancel) = builtin_yesno_prompt;
static const char *(*get_password)(const char *prompt) = builtin_get_password;
static int encrypted_channel = 0;

#ifndef USE_SHARED_LIBS
struct { const char *proto; tGPI fn; } static_protocols[] =
{
#ifdef PSERVER_PROTOCOL_SUPPORT	
	{ "pserver",	pserver_get_protocol_interface },
#endif
#ifdef SSERVER_PROTOCOL_SUPPORT	
	{ "sserver",	sserver_get_protocol_interface },
#endif
#ifdef GSERVER_PROTOCOL_SUPPORT	
	{ "gserver",	gserver_get_protocol_interface },
#endif
#ifdef SERVER_PROTOCOL_SUPPORT	
	{ "server",	server_get_protocol_interface },
#endif
#ifdef FORK_PROTOCOL_SUPPORT	
	{ "fork",	fork_get_protocol_interface },
#endif
#ifdef EXT_PROTOCOL_SUPPORT	
	{ "ext",	ext_get_protocol_interface },
#endif
#ifdef SSPI_PROTOCOL_SUPPORT	
	{ "sspi",	sspi_get_protocol_interface },
#endif
	{ NULL }
};
#else
static int initcount;
#endif

static int server_get_config_data(const struct server_interface *server, const char *key, const char *value, char *buffer, int buffer_len);
static int server_set_config_data(const struct server_interface *server, const char *key, const char *value, const char *buffer);
static int server_get_global_config_data(const struct server_interface *server, const char *key, const char *value, char *buffer, int buffer_len);
static int server_set_global_config_data(const struct server_interface *server, const char *key, const char *value, const char *buffer);
static int server_error(const struct server_interface *server, int fatal, const char *text);
static int server_getpass(char *password, int max_length, const char *prompt);
static int server_prompt(const char *message, const char *title, int withcancel);
static int server_set_encrypted_channel(int enctypt);

struct server_interface cvs_interface = 
{
	NULL, /* Current root */
	NULL, /* Library directory */
	NULL, /* Config directory */
	NULL, /* cvs command */
	0, /* Server input fd */
	1, /* Server output fd */
	
	server_get_config_data,
	server_set_config_data,
	server_get_global_config_data,
	server_set_global_config_data,
	server_error,
	server_getpass,
	server_prompt,
	server_set_encrypted_channel
};

struct server_interface *get_server_interface()
{
	return &cvs_interface;
}

const char *get_config_dir()
{
	return cvs_library_dir;
}

const char *get_library_dir()
{
	return cvs_config_dir;
}

bool set_config_dir(const char *dir)
{
	if(cvs_config_dir!=_cvs_config_dir_default)
		free((void*)cvs_config_dir);
	cvs_config_dir=strdup(dir);
	return true;
}

bool set_library_dir(const char *dir)
{
	if(cvs_library_dir!=_cvs_library_dir_default)
		free((void*)cvs_library_dir);
	cvs_library_dir=strdup(dir);
	return true;
}

void setup_server_interface(cvsroot_t *root, int io_socket, int (*_get_cached_password)(const char *key, char *buffer, int buffer_len),const char *(*_get_password)(const char *prompt),int (*_yesno_prompt)(const char *message, const char *title, int withcancel))
{
	CRunFile rf;
	cvs_interface.library_dir = cvs_library_dir;
	cvs_interface.config_dir = cvs_config_dir;
	cvs_interface.cvs_command = rf.program_name(); 
	cvs_interface.current_root = root;
	get_cached_password=_get_cached_password;
	get_password=_get_password?_get_password:builtin_get_password;
	yesno_prompt=_yesno_prompt?_yesno_prompt:builtin_yesno_prompt;
}

#ifdef USE_SHARED_LIBS
static int dlref()
{
   if(!initcount++)
	lt_dlinit();
}

static int dlunref()
{
   if(!--initcount)
      lt_dlexit();
}
#endif

const struct protocol_interface *load_protocol(const char *protocol)
{
#ifdef USE_SHARED_LIBS
	char fn[PATH_MAX];
	lt_dlhandle handle;
	tGPI get_protocol_interface;
	struct protocol_interface *proto_interface;

	if(!protocol)
		return NULL;

	snprintf(fn,sizeof(fn),"%s/%s_protocol",cvs_library_dir,protocol);

	CServerIo::trace(1,"Loading protocol %s",fn);

	dlref();	
	handle = lt_dlopenext(fn);

	if(!handle)
		return NULL; // Couldn't find protocol - not supported, or supporting DLLs missing

	get_protocol_interface = (tGPI)lt_dlsym(handle,"get_protocol_interface");
	if(!get_protocol_interface)
	{
		CServerIo::error("%s protocol library is corrupt",protocol);
		lt_dlclose(handle);
		dlunref();
		return NULL; // Couldn't find protocol - bad DLL
	}

	proto_interface = get_protocol_interface(&cvs_interface);
	proto_interface->__cvsnt_reserved=handle;
	proto_interface->name = strdup(protocol);
	return proto_interface;
#else /* static */
	struct protocol_interface *proto_interface;
	int n;

	CServerIo::trace(1,"Initialising protocol %s",protocol);

	for(n=0; static_protocols[n].proto; n++)
	{
	  if(!strcmp(protocol,static_protocols[n].proto))
	    break;
	}
	if(!static_protocols[n].proto)
	  return NULL;
	proto_interface = static_protocols[n].fn(&cvs_interface);
	proto_interface->name = strdup(protocol);
	return proto_interface;

#endif
}

int unload_protocol(const struct protocol_interface *protocol)
{
	if(protocol)
	{
		protocol->destroy(protocol);
		free((void*)((protocol_interface*)protocol)->name);
#ifdef USE_SHARED_LIBS
		lt_dlclose((lt_dlhandle)protocol->__cvsnt_reserved);
		dlunref();
#endif
	}
	return 0;
}

void get_config_file(const char *key, char *fn, int fnlen)
{
  struct passwd *pw = getpwuid(getuid());

  snprintf(fn,fnlen,"%s/.cvs",pw->pw_dir?pw->pw_dir:"");
  mkdir(fn,0777);
  snprintf(fn,fnlen,"%s/.cvs/%s",pw->pw_dir?pw->pw_dir:"",key?key:"");

  CServerIo::trace(2,"Config file name %s",fn);
}

int server_get_config_data(const struct server_interface *server, const char *key, const char *value, char *buffer, int buffer_len)
{
	char fn[PATH_MAX],line[1024];
	FILE *f;
	char *p;

	CServerIo::trace(3,"get_config-data(%s,%s,%d)",key,value,buffer_len);
	/* Special case for the 'cvspass' key */
	if(!strcmp(key,"cvspass") && get_cached_password && !get_cached_password(value,buffer,buffer_len))
	{
		CServerIo::trace(3,"returning cached value");
		return 0;
	}

	get_config_file(key,fn,sizeof(fn));
	f=fopen(fn,"r");
	if(f==NULL)
	  return -1;

	/* Read keys */
	while(fgets(line,sizeof(line),f))
	{
	  line[strlen(line)-1]='\0';
	  p=strchr(line,'=');
	  if(p)
	    *p='\0';
	  if(!strcasecmp(value,line))
	  {
	    if(p)
	      strncpy(buffer,p+1,buffer_len);
	    else
	      *buffer='\0';
	    return 0;
	  }
        }
	fclose(f);
	return -1;
}

int server_set_config_data(const struct server_interface *server, const char *key, const char *value, const char *buffer)
{
	char fn[PATH_MAX],fn2[PATH_MAX],line[1024];
	FILE *f,*o;
	char *p;
	int found;

	CServerIo::trace(3,"set_config_data(%s,%s)",key,value);

	get_config_file(key,fn,sizeof(fn));
	f=fopen(fn,"r");
	if(f==NULL)
	{
	  f=fopen(fn,"w");
	  if(f==NULL)
	  {
		CServerIo::trace(1,"Couldn't create config file %s",fn);
	    return -1;
	  }
	  if(buffer)
	    fprintf(f,"%s=%s\n",value,buffer);
	  fclose(f);
	  return 0;
	}
	sprintf(fn2,"%s.new",fn);
	o=fopen(fn2,"w");
	if(o==NULL)
	{
      CServerIo::trace(1,"Couldn't create temporary file %s",fn2);
	  fclose(f);
	  return -1;
	}

	/* Read keys */
	found=0;
	while(fgets(line,sizeof(line),f))
	{
	  line[strlen(line)-1]='\0';
	  p=strchr(line,'=');
	  if(p)
	    *p='\0';
	  if(!strcasecmp(value,line))
	  {
	    if(buffer)
	    {
	      strcat(line,"=");
	      strcat(line,buffer);
	      fprintf(o,"%s\n",line);
	    }
	    found=1;
	  }
	  else
	  {
	    if(p)
		*p='=';
	    fprintf(o,"%s\n",line);
	  }
        }
        if(!found && buffer)
	  fprintf(o,"%s=%s\n",value,buffer);
	fclose(f);
	fclose(o);
	rename(fn2,fn);
	return 0;
}

int server_error(const struct server_interface *server, int fatal, const char *text)
{
	/* If an auth module reports a fatal error, we need to log it 
           as the client probably won't see it */
#ifdef HAVE_SYSLOG_H
	syslog (LOG_DAEMON | (fatal?LOG_ERR:LOG_NOTICE), "%s", text);
#endif
	CServerIo::error("%s",text);
	if(fatal)
	   exit(-1);
	return 0;
}

#ifdef USE_SHARED_LIBS
static glob_t globbuf;
#endif

const char *enumerate_protocols(int *context)
{
#ifdef USE_SHARED_LIBS
	static char fn[PATH_MAX];
	char *p;

	if(!*context)
	{
		snprintf(fn,sizeof(fn),"%s/*_protocol.la",cvs_library_dir);
		globbuf.gl_offs = 0;
		if(glob(fn,GLOB_ERR|GLOB_NOSORT,NULL,&globbuf) != 0 || globbuf.gl_pathc == 0)
		  return NULL;
		*context=1;
	}
	else
	{
		(*context)++;
		if(*context>globbuf.gl_pathc)
		{
		  globfree(&globbuf);
		  *context=-1;
		  return NULL;
		}
	}
	strcpy(fn,globbuf.gl_pathv[(*context)-1]);
	*strrchr(fn,'_')='\0';
	p=strrchr(fn,'/');
	return p?p+1:fn; 
#else
	return static_protocols[(*context)++].proto;
#endif
}

void enumerate_protocols_end()
{
#ifdef USE_SHARED_LIBS
	globfree(&globbuf);
#endif
}	

const struct protocol_interface *find_authentication_mechanism(const char *tagline, bool& badauth,int io_socket, int (*_get_cached_password)(const char *key, char *buffer, int buffer_len),const char *(*_get_password)(const char *prompt),int (*_yesno_prompt)(const char *message, const char *title, int withcancel))
{
	int context;
	const char *proto;
	const struct protocol_interface *protocol;

	badauth=false;
	context=0;
	while((proto=enumerate_protocols(&context))!=NULL)
	{	
		protocol=load_protocol(proto);
		if(!protocol)
			continue;
		if(protocol->auth_protocol_connect)
		{
			int res;
                        setup_server_interface(NULL,io_socket,_get_cached_password,_get_password,_yesno_prompt);
			res = protocol->auth_protocol_connect(protocol,tagline);
			if(res==CVSPROTO_SUCCESS)
			{
				enumerate_protocols_end();
				return protocol; /* Correctly authenticated */
			}
			unload_protocol(protocol);
			if(res == CVSPROTO_AUTHFAIL)
			{
				badauth=true;
				return NULL;
			}
			if(res !=CVSPROTO_NOTME && res!=CVSPROTO_NOTIMP)
			{
				CServerIo::error("Authentication protocol rejected access");
				return NULL; /* Protocol was recognised, but failed */
			}
		}
		else
			unload_protocol(protocol);
	}
	return NULL;
}

void get_global_config_file(const char *key, char *fn, int fnlen)
{
  snprintf(fn,fnlen,"%s/%s",cvs_config_dir,key?key:"");
}

int server_get_global_config_data(const struct server_interface *server, const char *key, const char *value, char *buffer, int buffer_len)
{
	return get_global_config_data(key,value,buffer, buffer_len);
}

int get_global_config_data(const char *key, const char *value, char *buffer, int buffer_len)
{
        char fn[PATH_MAX],line[1024];
        FILE *f;
        char *p;

        get_global_config_file(key,fn,sizeof(fn));
        f=fopen(fn,"r");
        if(f==NULL)
          return -1;

        /* Read keys */
        while(fgets(line,sizeof(line),f))
        {
          line[strlen(line)-1]='\0';
          p=strchr(line,'=');
          if(p)
            *p='\0';
          if(!strcasecmp(value,line))
          {
            if(p)
              strncpy(buffer,p+1,buffer_len);
            else
              *buffer='\0';
            fclose(f);
            return 0;
          }
        }
        fclose(f);
        return -1;
}

int server_set_global_config_data(const struct server_interface *server, const char *key, const char *value, const char *buffer)
{
	return set_global_config_data(key,value,buffer);
}

int set_global_config_data(const char *key, const char *value, const char *buffer)
{
        char fn[PATH_MAX],fn2[PATH_MAX],line[1024];
        FILE *f,*o;
        char *p;
        int found;

        get_global_config_file(key,fn,sizeof(fn));
        f=fopen(fn,"r");
        if(f==NULL)
        {
          f=fopen(fn,"w");
          if(f==NULL)
          {
             CServerIo::trace(3,"Couldn't create config file %s\n",fn);
            return -1;
          }
          if(buffer)
            fprintf(f,"%s=%s\n",value,buffer);
          fclose(f);
          return 0;
        }
        sprintf(fn2,"%s.new",fn);
        o=fopen(fn2,"w");
        if(o==NULL)
        {
            CServerIo::trace(3,"Couldn't create temporary file %s\n",fn2);
          fclose(f);
          return -1;
        }

        /* Read keys */
        found=0;
        while(fgets(line,sizeof(line),f))
        {
          line[strlen(line)-1]='\0';
          p=strchr(line,'=');
          if(p)
            *p='\0';
          if(!strcasecmp(value,line))
          {
            if(buffer)
            {
              strcat(line,"=");
              strcat(line,buffer);
              fprintf(o,"%s\n",line);
            }
            found=1;
          }
	  else
	  {
	    if(p)
		*p='=';
	    fprintf(o,"%s\n",line);
	  }
        }
        if(!found && buffer)
          fprintf(o,"%s=%s\n",value,buffer);
        fclose(f);
        fclose(o);
        rename(fn2,fn);
        return 0;
}

int enum_global_config_data(const char *key, int value_num, char *value, int value_len, char *buffer, int buffer_len)
{
  char fn[PATH_MAX],line[1024];
  FILE *f;
  char *token,*v,*p;

  get_global_config_file(key,fn,sizeof(fn));
  f=fopen(fn,"r");
  if(f==NULL)
     return -1;

  /* Read keys */
  while(fgets(line,sizeof(line),f))
  {
    line[strlen(line)-1]='\0';
    if(line[0]=='#' || !strlen(line))
      continue;
    if(!value_num--)
    {
      for(token=line; isspace(*token); token++)
        ;
      v=p=strchr(token,'=');
      if(!p && !strlen(token))
        continue;
      if(p)
      {
        *p='\0';
        v++;
      }
      for(;isspace(*p); p++)
        *p='\0';
      for(;v && isspace(*v); v++)
        ;
      strncpy(value,token,value_len);
      if(p && v && strlen(v))
        strncpy(buffer,v,buffer_len);
      else
        *buffer='\0';
      fclose(f);
      return 0;
    }
  }
  fclose(f);
  return -1;
}

int server_getpass(char *password, int max_length, const char *prompt)
{
	const char *pw;
	if(get_password) pw = get_password(prompt);
	else pw=builtin_get_password(prompt);
	if(!pw)
		return 0;
	strncpy(password,pw,max_length);
	return 1;
}

const char *builtin_get_password(const char *prompt)
{
	return getpass(prompt);
}

int builtin_yesno_prompt(const char *message, const char *title, int withcancel)
{
	char c;

	fflush (stderr);
	fflush (stdout);
	fflush (stdin);

	printf("%s",message);
	fflush(stdout);
	for(;;)
	{
		c=getchar();
		if(tolower(c)=='y' || tolower(c)=='j' || tolower(c)=='o' || c=='\n' || c=='\r')
		{
			putchar(c);
			putchar('\n');
			fflush (stdin);
			return 1;
		}
		if(withcancel && (c==27 || tolower(c)=='c'))
		{
			putchar(c);
			putchar('\n');
			fflush (stdin);
			return -1;
		}
		if(tolower(c)=='n' || (!withcancel && c==27))
		{
			putchar(c);
			putchar('\n');
			fflush (stdin);
			return 0;
		}
	}
}

int server_prompt(const char *message, const char *title, int withcancel)
{
	return yesno_prompt(message,title,withcancel);
}

int server_set_encrypted_channel(int encrypt)
{
        set_encrypted_channel(encrypt);
        return 0;
}

void set_encrypted_channel(int encrypt)
{
	encrypted_channel = encrypt;
}

int get_encrypted_channel()
{
	return encrypted_channel;
}

#ifdef USE_SHARED_LIBS
void *library_load(const char *library, const char *function, library_handle_t& handle)
{
        void *proc;
        handle = NULL;
        cvs::string lib;

        if(CFileAccess::absolute(library) || CFileAccess::uplevel(library)>0)
        {
                CServerIo::error("Invalid library pathname '%s'",library);
                return NULL;
        }

	cvs::sprintf(lib,strlen(cvs_library_dir)+strlen(library)+10,"%s/%s",cvs_library_dir,library);

	dlref();
	handle = lt_dlopenext(lib.c_str());

	if(!handle)
	{
		dlunref();
		return NULL; // Couldn't find protocol - not supported, or supporting DLLs missing
	}

        if(!function)
                return NULL;

	proc = lt_dlsym((lt_dlhandle)handle,function);
        if(!proc)
        {
		lt_dlclose((lt_dlhandle)handle);
		dlunref();
		handle = NULL;
                return NULL;
        }
        return (void*)proc;
}

bool library_unload(library_handle_t library)
{
        if(!library)
                return true;
	lt_dlclose((lt_dlhandle)library);
	dlunref();
        return true;
}
#endif

