/* CVS auth library interface

    Copyright (C) 2004 Tony Hoyle and March-Hare Software Ltd

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
#define WIN32_LEAN_AND_MEAN
#define _WIN32_WINNT 0x0400
#include <windows.h>
#include <stdio.h>
#include <tchar.h>
#include "../windows-NT/config.h"
#include "../cvsapi/cvsapi.h"

#include "..\cvsservice\ServiceMsg.h"
void ReportError(BOOL bError, LPCSTR szError, ...);
void AddEventSource(LPCTSTR szService, LPCTSTR szModule);

#include "protocol_interface.h"
#include "library.h"

static int server_get_config_data(const struct server_interface *server, const char *key, const char *value, char *buffer, int buffer_len);
static int server_set_config_data(const struct server_interface *server, const char *key, const char *value, const char *buffer);
static int server_get_global_config_data(const struct server_interface *server, const char *key, const char *value, char *buffer, int buffer_len);
static int server_set_global_config_data(const struct server_interface *server, const char *key, const char *value, const char *buffer);
static int server_error(const struct server_interface *server, int fatal, const char *text);
static int server_getpass(char *password, int max_length, const char *prompt);
static int server_prompt(const char *message, const char *title, int withcancel);
static int server_set_encrypted_channel(int enctypt);
static int builtin_yesno_prompt(const char *message, const char *title, int withcancel);
static const char *builtin_get_password(const char *prompt);

static char cvs_dir[_MAX_PATH] = {0};
static char cvs_command[_MAX_PATH] = {0};
static int (*get_cached_password)(const char *key, char *buffer, int buffer_len);
static int (*yesno_prompt)(const char *message, const char *title, int withcancel) = builtin_yesno_prompt;
static const char *(*get_password)(const char *prompt) = builtin_get_password;
static int encrypted_channel = 0;

struct server_interface cvs_interface = 
{
	NULL, /* Current root */
	NULL, /* Library directory */
	NULL, /* Config directory */
	NULL, /* cvs command */
	0,	  /* Input FD */
	1,	  /* Output FD */
	
	server_get_config_data,
	server_set_config_data,
	server_get_global_config_data,
	server_set_global_config_data,
	server_error,
	server_getpass,
	server_prompt,
	server_set_encrypted_channel
};

static const char *get_cvs_dir()
{
	if(!*cvs_dir) /* Only do this once */
	{
		if (GetModuleFileNameA(NULL, cvs_dir, sizeof(cvs_dir)))
		{
			char *p;
			GetShortPathNameA(cvs_dir,cvs_dir,sizeof(cvs_dir));
			p = strrchr(cvs_dir, '\\');
			strcpy(cvs_command,cvs_dir);
			if(p)
				*p = '\0';
			else
				cvs_dir[0] = '\0';
		}
	}
	return cvs_dir;
}

const char *get_config_dir()
{
	return get_cvs_dir();
}

const char *get_library_dir()
{
	return get_cvs_dir();
}

bool set_config_dir(const char *dir)
{
	return false;
}

bool set_library_dir(const char *dir)
{
	return false;
}

struct server_interface *get_server_interface()
{
	return &cvs_interface;
}

void setup_server_interface(cvsroot_t *root, int io_socket, int (*_get_cached_password)(const char *key, char *buffer, int buffer_len),const char *(*_get_password)(const char *prompt),int (*_yesno_prompt)(const char *message, const char *title, int withcancel))
{
	get_cvs_dir();
	cvs_interface.library_dir = cvs_dir;
	cvs_interface.config_dir = cvs_dir;
	cvs_interface.cvs_command = cvs_command;
	cvs_interface.current_root = root;
	if(io_socket)
	{
		cvs_interface.in_fd = io_socket;
		cvs_interface.out_fd = io_socket;
	}
	get_cached_password=_get_cached_password;
	get_password=_get_password?_get_password:builtin_get_password;
	yesno_prompt=_yesno_prompt?_yesno_prompt:builtin_yesno_prompt;
}

const struct protocol_interface *load_protocol(const char *protocol)
{
	char fn[_MAX_PATH];
	HMODULE hLib;
	tGPI get_protocol_interface;
	struct protocol_interface *proto_interface;
	UINT oldMode;

	get_cvs_dir();
	_snprintf(fn,sizeof(fn),"%s/protocol_map.ini",cvs_dir);
	if(GetPrivateProfileStringA("cvsnt",protocol,NULL,fn,sizeof(fn),fn))
	{
		if(!stricmp(fn,"none") || !stricmp(fn,"off") || !stricmp(fn,"disable"))
			return NULL;
		if(!CFileAccess::absolute(fn))
		{
			char fn2[_MAX_PATH];
			_snprintf(fn2,sizeof(fn2),"%s/%s",cvs_dir,fn);
			strcpy(fn,fn2);
		}
	}
	else
		_snprintf(fn,sizeof(fn),"%s/%s_protocol.dll",cvs_dir,protocol);

    oldMode=SetErrorMode(SEM_NOOPENFILEERRORBOX|SEM_FAILCRITICALERRORS);
   	hLib = LoadLibraryA(fn);
	SetErrorMode(oldMode);

	if(!hLib)
		return NULL; // Couldn't find protocol - not supported, or supporting DLLs missing

	get_protocol_interface = (tGPI)GetProcAddress(hLib,"get_protocol_interface");
	if(!get_protocol_interface)
	{
		CServerIo::error("%s protocol library is corrupt",protocol);
		FreeLibrary(hLib);
		return NULL; // Couldn't find protocol - bad DLL
	}

	proto_interface = get_protocol_interface(&cvs_interface);
	proto_interface->__cvsnt_reserved=(void*)hLib;
	if(proto_interface->interface_version!=PROTOCOL_INTERFACE_VERSION)
	{
		CServerIo::trace(3,"Not loading %s - wrong version",protocol);
		if(proto_interface->destroy)
			proto_interface->destroy(proto_interface);
		FreeLibrary(hLib);
		return NULL;
	}
	proto_interface->name=strdup(protocol);
	return proto_interface;
}

int unload_protocol(const struct protocol_interface *protocol)
{
	if(protocol)
	{
		protocol->destroy(protocol);
		free((void*)((protocol_interface *)protocol)->name);
		FreeLibrary((HMODULE)protocol->__cvsnt_reserved);
	}
	return 0;
}

int server_get_config_data(const struct server_interface *server, const char *key, const char *value, char *buffer, int buffer_len)
{
	HKEY hKey,hSubKey;
	DWORD dwType,dwLen,dw;

	/* Special case for the 'cvspass' key */
	if(!strcmp(key,"cvspass") && get_cached_password && !get_cached_password(value,buffer,buffer_len))
			return 0;

	if(RegOpenKeyExA(HKEY_CURRENT_USER,"Software\\Cvsnt",0,KEY_READ,&hKey) &&
	   RegCreateKeyExA(HKEY_CURRENT_USER,"Software\\Cvsnt",0,NULL,0,KEY_READ,NULL,&hKey,NULL))
	{
		return -1; // Couldn't open or create key
	}

	if(key)
	{
		if(RegOpenKeyExA(hKey,key,0,KEY_READ,&hSubKey) &&
		   RegCreateKeyExA(hKey,key,0,NULL,0,KEY_READ,NULL,&hSubKey,NULL))
		{
			RegCloseKey(hKey);
			return -1; // Couldn't open or create key
		}
		RegCloseKey(hKey);
		hKey=hSubKey;
	}

	dwType=REG_SZ;
	dwLen=buffer_len;
	if((dw=RegQueryValueExA(hKey,value,NULL,&dwType,(LPBYTE)buffer,&dwLen))!=0)
	{
		RegCloseKey(hKey);
		return -1;
	}
	RegCloseKey(hKey);
	if(dwType==REG_DWORD)
		sprintf(buffer,"%u",*(DWORD*)buffer);

	return 0;
}

int server_set_config_data(const struct server_interface *server, const char *key, const char *value, const char *buffer)
{
	HKEY hKey,hSubKey;
	DWORD dwLen;

	if(RegOpenKeyExA(HKEY_CURRENT_USER,"Software\\Cvsnt",0,KEY_READ,&hKey) &&
	   RegCreateKeyExA(HKEY_CURRENT_USER,"Software\\Cvsnt",0,NULL,0,KEY_READ,NULL,&hKey,NULL))
	{
		return -1; // Couldn't open or create key
	}

	if(key)
	{
		if(RegOpenKeyExA(hKey,key,0,KEY_WRITE,&hSubKey) &&
		   RegCreateKeyExA(hKey,key,0,NULL,0,KEY_WRITE,NULL,&hSubKey,NULL))
		{
			RegCloseKey(hKey);
			return -1; // Couldn't open or create key
		}
		RegCloseKey(hKey);
		hKey=hSubKey;
	}

	if(!buffer)
	{
		RegDeleteValueA(hKey,value);
	}
	else
	{
		dwLen=(DWORD)strlen(buffer);
		if(RegSetValueExA(hKey,value,0,REG_SZ,(LPBYTE)buffer,dwLen+1))
		{
			RegCloseKey(hKey);
			return -1;
		}
	}
	RegCloseKey(hKey);

	return 0;
}

int server_error(const struct server_interface *server, int fatal, const char *text)
{
	ReportError(fatal,"%s",text);
	CServerIo::error("%s",text);
	return 0;
}

const char *enumerate_protocols(int *context)
{
	static HANDLE hFind=INVALID_HANDLE_VALUE;
	static char fn[_MAX_PATH];
	WIN32_FIND_DATAA wfd;

	if(!*context)
	{
		get_cvs_dir();
		_snprintf(fn,sizeof(fn),"%s/*_protocol.dll",cvs_dir);
		hFind=FindFirstFileA(fn,&wfd);
		if(hFind==INVALID_HANDLE_VALUE)
			return NULL;
		*context=1;
	}
	else
	{
		if(hFind==INVALID_HANDLE_VALUE)
			return NULL;
		if(!FindNextFileA(hFind,&wfd))
		{
			FindClose(hFind);
			hFind=INVALID_HANDLE_VALUE;
			*context=2;
			return NULL;
		}
	}
	strcpy(fn,wfd.cFileName);
	*strchr(fn,'_')='\0';
	return fn;
}

const struct protocol_interface *find_authentication_mechanism(const char *tagline, bool& badauth,int io_socket, int (*_get_cached_password)(const char *key, char *buffer, int buffer_len),const char *(*_get_password)(const char *prompt),int (*_yesno_prompt)(const char *message, const char *title, int withcancel))
{
	int context;
	const char *proto;
	const struct protocol_interface *protocol;
	int res;
	
	badauth=false;

	context=0;
	while((proto=enumerate_protocols(&context))!=NULL)
	{
		protocol=load_protocol(proto);
		if(!protocol)
			continue;
		if(protocol->auth_protocol_connect)
		{
			setup_server_interface(NULL,io_socket,_get_cached_password,_get_password,_yesno_prompt);
			res = protocol->auth_protocol_connect(protocol,tagline);
			if(res==CVSPROTO_SUCCESS)
				return protocol; /* Correctly authenticated */
			unload_protocol(protocol);
			if(res == CVSPROTO_AUTHFAIL)
			{
				badauth=true;
				return NULL;
			}
			if(res !=CVSPROTO_NOTME && res!=CVSPROTO_NOTIMP)
			{
				CServerIo::error("Authentication protocol rejected access");
				return NULL; /* Protocol was recognised, but failed */
			}
		}
	}
	return NULL;
}

int server_get_global_config_data(const struct server_interface *server, const char *key, const char *value, char *buffer, int buffer_len)
{
	return get_global_config_data(key,value,buffer, buffer_len);
}

int get_global_config_data(const char *key, const char *value, char *buffer, int buffer_len)
{
	HKEY hKey,hSubKey;
	DWORD dwType,dwLen;

	if(RegOpenKeyExA(HKEY_LOCAL_MACHINE,"Software\\CVS",0,KEY_READ,&hKey) &&
	   RegCreateKeyExA(HKEY_LOCAL_MACHINE,"Software\\CVS",0,NULL,0,KEY_READ,NULL,&hKey,NULL))
	{
		return -1; // Couldn't open or create key
	}

	if(key)
	{
		if(RegOpenKeyExA(hKey,key,0,KEY_READ,&hSubKey) &&
		   RegCreateKeyExA(hKey,key,0,NULL,0,KEY_READ,NULL,&hSubKey,NULL))
		{
			RegCloseKey(hKey);
			return -1; // Couldn't open or create key
		}
		RegCloseKey(hKey);
		hKey=hSubKey;
	}

	dwType=REG_SZ;
	dwLen=buffer_len;
	if(RegQueryValueExA(hKey,value,NULL,&dwType,(LPBYTE)buffer,&dwLen))
	{
		RegCloseKey(hKey);
		return -1;
	}
	RegCloseKey(hKey);
	if(dwType==REG_DWORD)
	    sprintf(buffer,"%u",*(DWORD*)buffer);

	return 0;
}

int server_set_global_config_data(const struct server_interface *server, const char *key, const char *value, const char *buffer)
{
	return set_global_config_data(key,value,buffer);
}

int set_global_config_data(const char *key, const char *value, const char *buffer)
{
	HKEY hKey,hSubKey;
	DWORD dwLen;

	if(RegOpenKeyExA(HKEY_LOCAL_MACHINE,"Software\\CVS",0,KEY_READ,&hKey) &&
	   RegCreateKeyExA(HKEY_LOCAL_MACHINE,"Software\\CVS",0,NULL,0,KEY_READ,NULL,&hKey,NULL))
	{
		return -1; // Couldn't open or create key
	}

	if(key)
	{
		if(RegOpenKeyExA(hKey,key,0,KEY_WRITE,&hSubKey) &&
		   RegCreateKeyExA(hKey,key,0,NULL,0,KEY_WRITE,NULL,&hSubKey,NULL))
		{
			RegCloseKey(hKey);
			return -1; // Couldn't open or create key
		}
		RegCloseKey(hKey);
		hKey=hSubKey;
	}

	if(!buffer)
	{
		RegDeleteValueA(hKey,value);
	}
	else
	{
		dwLen=(DWORD)strlen(buffer);
		if(RegSetValueExA(hKey,value,0,REG_SZ,(LPBYTE)buffer,dwLen+1))
		{
			RegCloseKey(hKey);
			return -1;
		}
	}
	RegCloseKey(hKey);

	return 0;
}

int enum_global_config_data(const char *key, int value_num, char *value, int value_len, char *buffer, int buffer_len)
{
	HKEY hKey,hSubKey;
	DWORD dwType,dwLen,dwValLen;
	DWORD dwRes;

	if(RegOpenKeyExA(HKEY_LOCAL_MACHINE,"Software\\CVS",0,KEY_READ,&hKey) &&
	   RegCreateKeyExA(HKEY_LOCAL_MACHINE,"Software\\CVS",0,NULL,0,KEY_READ,NULL,&hKey,NULL))
	{
		return -1; // Couldn't open or create key
	}

	if(key)
	{
		if(RegOpenKeyExA(hKey,key,0,KEY_READ,&hSubKey) &&
		   RegCreateKeyExA(hKey,key,0,NULL,0,KEY_READ,NULL,&hSubKey,NULL))
		{
			RegCloseKey(hKey);
			return -1; // Couldn't open or create key
		}
		RegCloseKey(hKey);
		hKey=hSubKey;
	}

	dwLen=buffer_len;
	dwValLen=value_len;
	if((dwRes=RegEnumValueA(hKey,value_num,value,&dwValLen,NULL,&dwType,(LPBYTE)buffer,&dwLen))!=0 && dwRes!=234)
	{
		RegCloseKey(hKey);
		return -1;
	}
	RegCloseKey(hKey);
	if(dwType==REG_DWORD)
	    sprintf(buffer,"%u",*(DWORD*)buffer);

	return 0;
}

static char win32_getkey()
{
  INPUT_RECORD buffer;
  DWORD EventsRead;
  char ch;
  HANDLE hInput = GetStdHandle(STD_INPUT_HANDLE);
  BOOL CharRead = FALSE;

  /* loop until we find a valid keystroke (a KeyDown event, and NOT a
     SHIFT, ALT, or CONTROL keypress by itself) */
  while(!CharRead)
  {
	while( !CharRead && ReadConsoleInput(hInput, &buffer, 1, &EventsRead ) &&
			EventsRead > 0 )
	{
		if( buffer.EventType == KEY_EVENT &&
			buffer.Event.KeyEvent.bKeyDown )
		{
			ch = buffer.Event.KeyEvent.uChar.AsciiChar;
			if(ch)
				CharRead = TRUE;
		}
	}
  }

  return ch;
}

/* This is normally used by cvsgui */
char *wnt_getpass (const char *prompt)
{
    static char password[128];
	static const int max_length = sizeof(password);
    int i;
	char c;
	HANDLE hInput=GetStdHandle(STD_INPUT_HANDLE);
	DWORD dwMode;

    fputs (prompt, stderr);
    fflush (stderr);
    fflush (stdout);
	FlushConsoleInputBuffer(hInput);
	GetConsoleMode(hInput,&dwMode);
	SetConsoleMode(hInput,ENABLE_PROCESSED_INPUT);
    for (i = 0; i < max_length - 1;)
    {
		c=0;
		c = win32_getkey();
		if(c==27 || c=='\r' || c=='\n')
			break;
		else if(c==8 && i)
		{
		  fputs("\b \b",stdout);
		  --i;
		}
		else if(c>31)
		{
		  password[i++]=c;
		  fputc('*',stdout);
		}
		fflush (stderr);
		fflush (stdout);
    }
	SetConsoleMode(hInput,dwMode);
	FlushConsoleInputBuffer(hInput);
    password[i] = '\0';
    fputs ("\n", stderr);
	return c==27?NULL:password;
}

int server_getpass(char *password, int max_length, const char *prompt)
{
	const char *pw;
	if(get_password) pw = get_password(prompt);
	else pw=builtin_get_password(prompt);

	if(!pw)
		return 0;
	strncpy(password,pw,max_length);
	return 1;
}

const char *builtin_get_password(const char *prompt)
{
	return wnt_getpass(prompt);
}

int builtin_yesno_prompt(const char *message, const char *title, int withcancel)
{
	char c;

	fflush (stderr);
	fflush (stdout);
	fflush (stdin);

	printf("%s",message);
	fflush(stdout);
	for(;;)
	{
		c=getchar();
		if(tolower(c)=='y' || tolower(c)=='j' || tolower(c)=='o' || c=='\n' || c=='\r')
		{
			putchar(c);
			putchar('\n');
			fflush (stdin);
			return 1;
		}
		if(withcancel && (c==27 || tolower(c)=='c'))
		{
			putchar(c);
			putchar('\n');
			fflush (stdin);
			return -1;
		}
		if(tolower(c)=='n' || (!withcancel && c==27))
		{
			putchar(c);
			putchar('\n');
			fflush (stdin);
			return 0;
		}
	}
}

int server_prompt(const char *message, const char *title, int withcancel)
{
	return yesno_prompt(message,title,withcancel);
}

int server_set_encrypted_channel(int encrypt)
{
        set_encrypted_channel(encrypt);
        return 0;
}

void set_encrypted_channel(int encrypt)
{
	encrypted_channel = encrypt;
}

int get_encrypted_channel()
{
	return encrypted_channel;
}

void ReportError(BOOL bError, LPCSTR szError, ...)
{
	static BOOL bEventSourceAdded = FALSE;
	char buf[512];
	const char *bufp = buf;
	va_list va;
	HANDLE hEvent;

	va_start(va,szError);
	vsprintf(buf,szError,va);
	va_end(va);
	if(!bEventSourceAdded)
	{
		TCHAR szModule[MAX_PATH];
		GetModuleFileName(NULL,szModule,MAX_PATH);
		AddEventSource(_T("cvsnt"),szModule);
		bEventSourceAdded=TRUE;
	}

	hEvent = RegisterEventSource(NULL,  _T("cvsnt"));
	ReportEventA(hEvent,bError?EVENTLOG_ERROR_TYPE:EVENTLOG_INFORMATION_TYPE,0,MSG_STRING,NULL,1,0,&bufp,NULL);
	DeregisterEventSource(hEvent);
}

void AddEventSource(LPCTSTR szService, LPCTSTR szModule)
{
	HKEY hk;
	DWORD dwData;
	TCHAR szKey[1024];

	_tcscpy(szKey,_T("SYSTEM\\CurrentControlSet\\Services\\EventLog\\Application\\"));
	_tcscat(szKey,szService);

    // Add your source name as a subkey under the Application 
    // key in the EventLog registry key.  
    if (RegCreateKey(HKEY_LOCAL_MACHINE, szKey, &hk))
		return; // Fatal error, no key and no way of reporting the error!!!

    // Add the name to the EventMessageFile subkey.  
    if (RegSetValueEx(hk,             // subkey handle 
            _T("EventMessageFile"),       // value name 
            0,                        // must be zero 
            REG_EXPAND_SZ,            // value type 
            (LPBYTE) szModule,           // pointer to value data 
            (DWORD)_tcslen(szModule)*sizeof(TCHAR) + sizeof(TCHAR)))       // length of value data 
			return; // Couldn't set key

    // Set the supported event types in the TypesSupported subkey.  
    dwData = EVENTLOG_ERROR_TYPE | EVENTLOG_WARNING_TYPE | 
        EVENTLOG_INFORMATION_TYPE;  
    if (RegSetValueEx(hk,      // subkey handle 
            _T("TypesSupported"),  // value name 
            0,                 // must be zero 
            REG_DWORD,         // value type 
            (LPBYTE) &dwData,  // pointer to value data 
            sizeof(DWORD)))    // length of value data 
			return; // Couldn't set key
	RegCloseKey(hk); 
} 

void *library_load(const char *library, const char *function, library_handle_t& handle)
{
	PROC proc;
	handle = NULL;
	cvs::string lib;
	UINT oldMode;

	if(CFileAccess::absolute(library) || CFileAccess::uplevel(library)>0)
	{
		CServerIo::error("Invalid library pathname '%s'",library);
		return NULL;
	}

	get_cvs_dir();
	cvs::sprintf(lib,strlen(library)+strlen(cvs_dir)+10,"%s/%s",cvs_dir,library);
    oldMode=SetErrorMode(SEM_NOOPENFILEERRORBOX|SEM_FAILCRITICALERRORS);
	handle = (HMODULE)LoadLibraryA(library);
	SetErrorMode(oldMode);
	if(!function)
		return NULL;
	proc=GetProcAddress((HMODULE)handle,function);
	if(!proc)
	{
		FreeLibrary((HMODULE)handle);
		handle = NULL;
		return NULL;
	}
	return (void*)proc;
}

bool library_unload(library_handle_t library)
{
	if(!library)
		return true;
	FreeLibrary((HMODULE)library);
	return true;
}
