/* run.c --- routines for executing subprocesses.
   
   This file is part of GNU CVS.

   GNU CVS is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the
   Free Software Foundation; either version 2, or (at your option) any
   later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.  */

#include "cvs.h"

#include <signal.h>

#ifndef HAVE_UNISTD_H
extern int execvp(char *file, char **argv);
#endif

static pid_t pipe_pid;

/*
 * To exec a program under CVS, first call run_setup() to setup initial
 * arguments.  The argument to run_setup will be parsed into whitespace 
 * separated words and added to the global run_argv list.
 * 
 * Then, optionally call run_arg() for each additional argument that you'd like
 * to pass to the executed program.
 * 
 * Finally, call run_exec() to execute the program with the specified arguments.
 * The execvp() syscall will be used, so that the PATH is searched correctly.
 * File redirections can be performed in the call to run_exec().
 */
static char **run_argv;
static int run_argc;
static int run_argc_allocated;

void run_setup (const char *prog)
{
    int i;

    /* clean out any malloc'ed values from run_argv */
    for (i = 0; i < run_argc; i++)
    {
		if (run_argv[i])
			xfree (run_argv[i]);
    }
    run_argc = 0;

    /* put each word into run_argv, allocating it as we go */
	run_arg (prog);
}

void run_arg (const char *s)
{
    TRACE(3,"run_arg %s",PATCH_NULL(s));
	/* allocate more argv entries if we've run out */
    if (run_argc >= run_argc_allocated)
    {
        run_argc_allocated += 50;
        run_argv = (char **) xrealloc ((char *) run_argv,
                        run_argc_allocated * sizeof (char **));
    }

    if (s)
    {
        shell_escape(&run_argv[run_argc],s);
        run_argc++;
    }
    else
        run_argv[run_argc] = (char *) 0;        /* not post-incremented on purpose! */
}

int run_exec (int flags)
{
  pid_t pid;
  int status;
  
  cvs_flusherr();
  cvs_flushout();
  run_arg(NULL);

#ifdef SERVER_SUPPORT
  if(server_active)
  {
    int stdout_pipe[2];
    int stderr_pipe[2];
    int stdout_fd, stderr_fd;
    int size;
    char buf[BUFSIZ];
    int w;

    if (pipe (stdout_pipe) < 0)
        error (1, 0, "cannot create pipe");
    if (pipe (stderr_pipe) < 0)
        error (1, 0, "cannot create pipe");

#ifdef USE_SETMODE_BINARY
    setmode (stdout_pipe[0], O_BINARY);
    setmode (stdout_pipe[1], O_BINARY);
    setmode (stderr_pipe[0], O_BINARY);
    setmode (stderr_pipe[1], O_BINARY);
#endif

#ifdef HAVE_VFORK
    pid = vfork ();
#else
    pid = fork ();
#endif
    if (pid < 0)
        error (1, 0, "cannot fork");
    if (pid == 0)
    {
        if (dup2 (stdout_pipe[1], 1) < 0)
          error (1, 0, "cannot dup2 pipe");
        if (close (stdout_pipe[0]) < 0)
          error (1, 0, "cannot close pipe");
        if (dup2 (stderr_pipe[1], 2) < 0)
          error (1, 0, "cannot dup2 pipe");
        if (close (stderr_pipe[0]) < 0)
          error (1, 0, "cannot close pipe");

        execvp (run_argv[0], run_argv);
        error (1, errno, "cannot exec %s", run_argv[0]);
    }
    if (close (stdout_pipe[1]) < 0)
      error (1, 0, "cannot close pipe");
    if (close (stderr_pipe[1]) < 0)
      error (1, 0, "cannot close pipe");

    stdout_fd = stdout_pipe[0];
    stderr_fd = stderr_pipe[0];

    fcntl(stdout_fd,F_SETFL, O_NONBLOCK);
    fcntl(stderr_fd,F_SETFL, O_NONBLOCK);
    do
    {
      status = 0;
      w = waitpid(pid, &status, WNOHANG); 
      usleep(100);
      while(stderr_fd && (size=read(stderr_fd,buf,BUFSIZ))>0)
        cvs_outerr(buf,size);
      if(size<0 && errno!=EAGAIN)
      {
        close(stderr_fd);
        stderr_fd = 0;
      }
      while(stdout_fd && (size=read(stdout_fd,buf,BUFSIZ))>0)
        cvs_output(buf,size);
      if(size<0 && errno!=EAGAIN)
      {
        close(stdout_fd);
        stdout_fd = 0;
      }
    } while(w==0);
  }
  else
#endif
  {
#ifdef HAVE_VFORK
     pid = vfork ();
#else
     pid = fork ();
#endif
    if (pid < 0)
        error (1, 0, "cannot fork");
     if(pid == 0)
     {
       execvp(run_argv[0],run_argv);
       error (1, errno, "cannot exec %s", run_argv[0]);
     }
     waitpid(pid, &status, 0);
  }

  cvs_flusherr();
  cvs_flushout();

  return WEXITSTATUS(status);
}

void run_print (FILE *fp)
{
    int i;
    int (*outfn)(const char *, size_t);

    if (fp == stderr)
	outfn = cvs_outerr;
    else if (fp == stdout)
	outfn = cvs_output;
    else
    {
	error (1, 0, "internal error: bad argument to run_print");
	/* Solely to placate gcc -Wall.
	   FIXME: it'd be better to use a function named `fatal' that
	   is known never to return.  Then kludges wouldn't be necessary.  */
	outfn = NULL;
    }

    for (i = 0; i < run_argc; i++)
    {
	(*outfn) ("'", 1);
	(*outfn) (run_argv[i], 0);
	(*outfn) ("'", 1);
	if (i != run_argc - 1)
	    (*outfn) (" ", 1);
    }
}

void close_on_exec (int fd)
{
#ifdef F_SETFD
    if (fcntl (fd, F_SETFD, 1))
	error (1, errno, "can't set close-on-exec flag on %d", fd);
#endif
}
